/*
 *****************
 *
 * Array List
 *
 *****************
 *
 * Estrucutra de datos que implementa un arreglo de tamaño dinamico
 *
 * La estructura es generica, pueto que el dato es un puntero a void
 *
 * para usar la busqueda y la funnciona "add" esta estrucutra requiere
 * de una funcion que sepa como comparar (int compareTo(void*,void*))
 * los objetos que se mantienen en la estructura
 *
 * 25-sep-2015
 * ortigosa
 *
 */
#ifndef ARRAY_LIST_H
#define ARRAY_LIST_H 1

typedef struct _ArrayList {
    int vectorSize;
    int numElements;
    void** vector;
    int (*compareTo)(void*,void*);
    int defaultVecSize;
} ArrayList;

typedef ArrayList Stack;

// crea una nuevo ArrayList
ArrayList* new_ArrayList();

// debe de usarse este vector al instancias un objeto de este tipo
// cambiar tamaño del arreglo que por default es VECTOR_SIZE
// elimina los el vector anterior
void ArrayList_setDefaultVectorSize(ArrayList* l, int size); 

// checa si la lista esta vacia
int ArrayList_isEmpty(ArrayList* l);

// regresa el numero de elementos en la lista
int ArrayList_getSize(ArrayList* l);

// establecer funcion comparadora de elementos de la lista
void ArrayList_setComparator(ArrayList* l, int (*compareTo)(void*, void*));

// agrega elementos al final de la lista
void ArrayList_add(ArrayList* l, void* data);

// agrega un elemento a la lista en forma ordenada
// esta funcion requiere que el objeto ArrayList sepa como comparar
// los objetos que contiene, es decir que implemente la funcion compareTo
void ArrayList_insert(ArrayList* l, void* data);

// mueve todos los elementos a la derecha dejando un hueco en i
// en caso de desborde duplica el tamaño del vector
void recorrerElementos(ArrayList* l, int i);

// crea un vector el doble de grande que el anterio 
void** getDoubleVector(ArrayList* l);

// obtiene un elemento de la lista dada su posicion
void* ArrayList_get(ArrayList* l, int index);

// pone un dato en una posicion especifica
void ArrayList_set(ArrayList* l, int index, void* data);

// busca un elemento en el arreglo
void* ArrayList_find(ArrayList* l, void* data);

// implementa busqueda binaria recursiva
int ArrayList_bb(ArrayList* l, int i, int j, void* data);

// indica si el elemento se encuentra o no
int ArrayList_contains(ArrayList* l, void* data);

// libera la memoria de una lista
void ArrayList_free(ArrayList* l);

// checa si se implemento la funcion para comparar elementos de la lista
void ArrayList_isCompareToDefined(ArrayList* l);

// elimina un elemento de la lista
void ArrayList_remove(ArrayList* l, void* data);

// recorre los elementos hacia la izquierda para cubir un hueco
void recorrerElementosIzquierda(ArrayList* l, int i);

// print list
void ArrayList_print(ArrayList* l, void(*print)(void*));

// indica si un elementos se encuentra en la lista dado un comparador
int ArrayList_contiene(ArrayList* l, void* data, int (*eq)(void*,void*));

//-------------------------------------------------
// funciones para comportamiento de pila
//-------------------------------------------------
//


Stack* new_Stack();

void* Stack_pop(Stack* s);

void* Stack_peek(Stack* s);

void Stack_push(Stack* s, void* data);

int Stack_isEmpty(Stack* s);

void Stack_setComparator(Stack* s, int (*cmp)(void*, void*));

#endif

