/*
 *****************
 *
 * ArrayList
 *
 *****************
 *
 * Estrucutra de datos que implementa un arreglo de tamaño dinamico
 *
 * La estructura es generica, pueto que el dato es un puntero a void
 *
 * para usar la busqueda y la funnciona "add" esta estrucutra requiere
 * de una funcion que sepa como comparar (int compareTo(void*,void*))
 * los objetos que se mantienen en la estructura
 *
 * Ortigosa Flores Jose Luis
 * 25-sep-2015
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/ArrayList.h"

#define VECTOR_SIZE 2000

/*

typedef struct _ArrayList {
    int vectorSize;
    int numElements;
    void** vector;
    int (*compareTo)(void*,void*);
} ArrayList;

*/

// crea una nuevo ArrayList
ArrayList* new_ArrayList() {
    ArrayList* n = (ArrayList*) malloc(sizeof(ArrayList));
    if(n == NULL) {
	exit(-1);
    }
    n->defaultVecSize = VECTOR_SIZE;
    n->vectorSize = VECTOR_SIZE;
    n->numElements = 0;
    n->vector = (void*) calloc(VECTOR_SIZE, sizeof(void*));
    n->compareTo = NULL;
    return n;
}

// debe de usarse este vector al instancias un objeto de este tipo
// cambiar tamaño del arreglo que por default es VECTOR_SIZE
// elimina los el vector anterior
void ArrayList_setDefaultVectorSize(ArrayList* l, int size) {
    l->defaultVecSize = size;
    l->vectorSize = size;
    l->numElements = 0;
    free(l->vector);
    l->vector = (void*) calloc(size, sizeof(void*));
}

// checa si la lista esta vacia
int ArrayList_isEmpty(ArrayList* l) {
    return l->numElements == 0;
}

// regresa el numero de elementos en la lista
int ArrayList_getSize(ArrayList* l) {
    return l->numElements;
}

// establecer funcion comparadora de elementos de la lista
void ArrayList_setComparator(ArrayList* l, int (*compareTo)(void*, void*)) {
    l->compareTo = compareTo;
}


// mueve todos los elementos a la derecha dejando un hueco en i
// en caso de desborde duplica el tamaño del vector
void recorrerElementos(ArrayList* l, int i) {
    int k;
    for(k=l->numElements; k>i; k--) {
	l->vector[k] = l->vector[k-1];
    }
}

// obtiene un vector del doble del tamaño de la lista
void doubleVector(ArrayList* l) {
    void* doubleVec = calloc(2*l->vectorSize, sizeof(void*));
    memcpy(doubleVec, l->vector, l->numElements*sizeof(void*));
    free(l->vector);
    l->vector = doubleVec;
    l->vectorSize=2*l->vectorSize;
}

// agrega un elemento a la lista en forma ordenada
// esta funcion requiere que el objeto ArrayList sepa como comparar
// los objetos que contiene, es decir que implemente la funcion compareTo
void ArrayList_insert(ArrayList* l, void* data) {
    ArrayList_isCompareToDefined(l);

    if(ArrayList_isEmpty(l)) {
	l->vector[0] = data;
	l->numElements++;
	return;
    }

    
    if(l->numElements+1 >= l->vectorSize) {
	doubleVector(l);
    }

    int i;
    for(i=0; i < l->numElements; i++) {
        if(l->compareTo(data, l->vector[i]) < 0) {
	    recorrerElementos(l, i);
	    l->vector[i] = data;
	    l->numElements++;
	    return;
	}
    }

    if(i == l->numElements) {
	l->vector[i] = data;
	l->numElements++;
    }
    
}

// agrega elementos al final de la lista
void ArrayList_add(ArrayList* l, void* data) {
    if(l->numElements+1 >= l->vectorSize) {
	doubleVector(l);
    }
    l->vector[l->numElements++] = data;
}


// obtiene un elemento de la lista dada su posicion
void* ArrayList_get(ArrayList* l, int index) {
    if(index >= 0 || index < l->numElements) {
	return l->vector[index];
    }
}

// pone un dato en una posicion especifica
void ArrayList_set(ArrayList* l, int index, void* data) {
    if(index >= 0 || index < l->numElements) {
	l->vector[index] = data;
    }
}

    
// busca un elemento en el arreglo
void* ArrayList_find(ArrayList* l, void* data) {
    ArrayList_isCompareToDefined(l);
    int index = ArrayList_bb(l, 0, l->numElements-1, data);
    if(index == -1) return NULL;
    return ArrayList_get(l, index);
}

// implementa busqueda binaria recursiva
int ArrayList_bb(ArrayList* l, int i, int j, void* data) {
    
    if(i <= j) {
	int m = ((j-i)/2) + i;
	if(l->compareTo(data, l->vector[m]) == 0) {
	    return m;
	} else if(l->compareTo(data, l->vector[m]) < 0) {
	    return ArrayList_bb(l, i, m-1, data);
	} else {
	    return ArrayList_bb(l, m+1, j, data);
	}
    }
    return -1;
}

// indica si el elemento se encuentra o no
int ArrayList_contains(ArrayList* l, void* data) {
    ArrayList_isCompareToDefined(l);
    int cont = ArrayList_bb(l, 0, l->numElements-1, data);
    return cont > -1;
}

// libera la memoria de una lista
void ArrayList_free(ArrayList* l) {
    free(l->vector);
    free(l);
}

// checa si se implemento la funcion para comparar elementos de la lista
void ArrayList_isCompareToDefined(ArrayList* l) {
    if(l->compareTo == NULL) {
	perror("Error: no se definio funcion: int compareTo(void*,void*)\n");
	exit(-1);
    }
}

// elimina un elemento de la lista
void ArrayList_remove(ArrayList* l, void* data) {
    ArrayList_isCompareToDefined(l);
    int index = ArrayList_bb(l, 0, l->numElements-1, data);
    if(index != -1) {
	recorrerElementosIzquierda(l, index);
	l->numElements--;
    }
}

// recorre los elementos hacia la izquierda para cubir un hueco
void recorrerElementosIzquierda(ArrayList* l, int i) {
    void* ptr = l->vector[i];
    int j;
    for(j=i; j<l->numElements-1; j++) {
	l->vector[j] = l->vector[j+1];
    }
    free(ptr);
}

// print list
void ArrayList_print(ArrayList* l, void(*print)(void*)) {
    printf("List size: %d\n ", l->numElements);
    int i;
    for(i=0; i<l->numElements; i++) {
	printf(" ");
	print(l->vector[i]);
    }
}

int ArrayList_contiene(ArrayList* l, void* data, int (*eq)(void*,void*)) {
    int i;
    void* d;
    for(i=0; i<l->numElements; i++) {
	    d = l->vector[i];
	    if(eq(d, data)) {
	        return 1;
	    }
    }
    return 0;
}

int ArrayList_del(ArrayList* l, void* data, int (*eq)(void*,void*)) {
    int i;
    for(i=0; i<l->numElements; i++) {
        if(eq(data, l->vector[i])) {
            recorrerElementosIzquierda(l, i);
            l->numElements--;
            return ;
        }
    }
}

//-------------------------------------------------
// funciones para comportamiento de pila
//-------------------------------------------------
//

Stack* new_Stack() {
    return new_ArrayList();
}

void* Stack_pop(Stack* s) {
    void* data = Stack_peek(s);
    s->numElements--;
    return data;
}

void* Stack_peek(Stack* s) {
    return ArrayList_get(s, ArrayList_getSize(s)-1);
}

void Stack_push(Stack* s, void* data) {
    ArrayList_add(s, data);
}

int Stack_isEmpty(Stack* s) {
    return ArrayList_isEmpty(s);
}

void Stack_setComparator(Stack* s, int (*cmp)(void*, void*)) {
    ArrayList_setComparator(s,cmp);
}


