/*
 *****************************
 * 
 * Forma
 *
 *****************************
 * 
 * Objeto que almacena los puntos que conforman 
 * una forma que se encuentra una imagen
 *
 *
 * Jose Luis Ortigosa Flores
 * 25-sep-2015
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include "../include/ArrayList.h"
#include "../include/Punto.h"
#include "../include/Forma.h"

/*
typedef struct _Forma {
    ArrayList* points;
    Punto min;
    Punto max;
} Forma;
*/

#define MAX_PX 10000000
#define MIN_PX 0

Forma* new_Forma() {
    Forma* f = (Forma*) malloc(sizeof(Forma));
    if(f == NULL) {
	    exit(-1);
    }

    f->vertices = new_ArrayList();
    ArrayList_setDefaultVectorSize(f->vertices, 10);
    f->min.x = MAX_PX;
    f->min.y = MAX_PX;
    f->max.x = MIN_PX;
    f->max.y = MIN_PX;
    f->numPoints = 0;
    f->var = NULL;
    return f;
}

// agrega un punto a la forma
void Forma_addPoint(Forma* f, Punto* p) {
    actualizarPuntoMax(f, p);
    actualizarPuntoMin(f, p);
    f->numPoints++;
}

// checa y actualiza si el punto p es major
void actualizarPuntoMax(Forma* f, Punto* p) {
    if(Punto_cmpX(p, &f->max) > 0) {
	    f->max.x = p->x;
    }
    if(Punto_cmpY(p, &f->max) > 0) {
	    f->max.y = p->y;
    }
}

// checa y actualiza si el punto p es menor
void actualizarPuntoMin(Forma* f, Punto* p) {
    if(Punto_cmpX(p, &f->min) < 0) {
	    f->min.x = p->x;
    }
    if(Punto_cmpY(p, &f->min) < 0) {
	    f->min.y = p->y;
    }
}


// regresa la cantidad de puntos que la conforman
int Forma_getNumPoints(Forma* f) {
    return f->numPoints;
}

// obtiene el punto maximo dentro de la forma
Punto* Forma_getMaxPoint(Forma* f) {
    return &f->max;
}

// obtiene el punto minimo dentro de la forma
Punto* Forma_getMinPoint(Forma* f) {
    return &f->min;
}

//libera la memoria
void Forma_free(Forma* f) {
    ArrayList_free(f->vertices);
    free(f);
}


// print forma
void Forma_print(void* o) {
    Forma* f = (Forma*)o;
    int np = Forma_getNumPoints(f);
    printf("\n# points: %d", np);
    printf("\nMin point: ");
    Punto* minP = Forma_getMinPoint(f);
    Punto_print(minP);
    printf("\nMax point: ");
    Punto* maxP = Forma_getMaxPoint(f);
    Punto_print(maxP);
    printf("\n");
}

ArrayList* Forma_getVertices(Forma* f) {
    return f->vertices;
}

// Regresa una lista de vertices
ArrayList* Forma_escogerVertices(ArrayList* vertices) {
    Punto* amax;
    Punto* bmax;
    Punto* cmax;
    double areaMax = -1000000, area;
    Punto* a;
    Punto* b;
    Punto* c;
    Punto* r1;
    Punto* r2;
    int i,j,k, size = ArrayList_getSize(vertices);

    for(i=0; i<size; i++) {
        for(j=0; j<size; j++) {
            for(k=0; k<size; k++) {
                if(i!=j && j!=k) {
                    a = ArrayList_get(vertices, i);
                    b = ArrayList_get(vertices, j);
                    c = ArrayList_get(vertices, k);

                    r1 = Punto_restarAbs(a, b);
                    r2 = Punto_restarAbs(a, c);

                    area = Punto_area(r1, r2);

                    if(area<0) {
                        area *= -1;
                    }

                    if(area > areaMax) {
                        amax = a;
                        bmax = b;
                        cmax = c;
                        areaMax = area;
                    }
                }
            }
        }
    }

    ArrayList* l = new_ArrayList();
    ArrayList_add(l, amax);
    ArrayList_add(l, bmax);
    ArrayList_add(l, cmax);
    return l;
}

// Agrega un vertice
void Forma_addVertice(Forma* f, Punto* p) {
    if(!ArrayList_contiene(f->vertices, p, Punto_equals)) {
	    ArrayList_add(f->vertices, p);
    }
}

void Forma_setVerticeAnguloRecto(Forma* f, Punto* p) {
    f->var = p;
}

// regresa el vertice con en el que se forma el angulo recto
Punto* Forma_getVerticeAnguloRecto(Forma* f) {
    return f->var;
}


