/*
 *********************************
 * 
 * CINVESTAV-IPN
 * Realidad Aumentada
 *
 * Tarea 2 Reconocimiento de Triangulos
 *
 * Jose Luis Ortigosa Flores
 * 28-sep-15
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include "../include/ArrayList.h"
#include "../include/Punto.h"
#include "../include/Forma.h"
#include "../../software/Scimagen/matrix.h"
#include "../../software/Scimagen/pnm.h"

#define MAX_CHARS 64

const int COLOR_GRAY = 220;
const int COLOR_DARK_GRAY = 120;
int numForma = 1;



void buscarFormas(MATRIX* m, ArrayList* formas);
Forma* obtenerForma(MATRIX* m, int i, int j);
void agregarVecinos(Punto* p, Stack* s, MATRIX* m);
void crearCajaEnvolvente(MATRIX* m, Punto* min, Punto* max);
int esVertice(Punto* p, MATRIX* m);
int esEsquinaIzqArriba(Punto* p, MATRIX* m);
int esEsquinaDerArriba(Punto* p, MATRIX* m);
int esEsquinaDerAbajo(Punto* p, MATRIX* m);
int esEsquinaIzqAbajo(Punto* p, MATRIX* m);
void dibujarVertices(MATRIX* m, ArrayList* v);
void dibujarPunto(MATRIX* m, Punto* p, int tonoGris, int grosor);
Punto* buscarAnguloRecto(ArrayList* v);
void dibujarNumero(MATRIX* m, Punto* p, int n, int tonoGris);
void analisisImagen(ArrayList* formas, char* name);


int main(int argc, char** argv) {
    if(argc != 3) {
        printf("recTriangulos <imagen a analizar> <imagen salida>");
    }
    
    MATRIX* imgIn;
    char* imgInName = argv[1];
    char* imgOutName = argv[2];
    // lista para alamacenar las formas encontradas en la imagen
    ArrayList* formas = new_ArrayList();

    // cargo la imagen en memoria
    imgIn = read_pnm(imgInName);

    // busco en toda la imagen todas las formas posibles y las almaceno
    // en la lista
    buscarFormas(imgIn, formas);

    Forma* f;
    ArrayList* v;
    Punto* ar;
    int i;
    for(i=0; i<ArrayList_getSize(formas); i++) {

        // para cada forma encontrada en la imagen
	    f = ArrayList_get(formas, i);

        // creo la caja envolvente
        crearCajaEnvolvente(imgIn, &f->min, &f->max);
	    
        // que fueron encontrados durante la busqueda de formas
        v = Forma_getVertices(f);
       
        // obtenogo los vertices
        v = Forma_escogerVertices(v);
        f->vertices = v;
        
        // dibujo los vertices en la imagen
	    dibujarVertices(imgIn, v);

        // busco los angulos rectos dado un conjunto de vertoces
	    ar = buscarAnguloRecto(v);
        Forma_setVerticeAnguloRecto(f, ar);

        // dibujo en la imagen el angulo recto 
        dibujarPunto(imgIn, ar, 100, 4);

        // dibujo a cada forma su etiqueta
        dibujarNumero(imgIn, &f->min, i+1, 50);

    }

    //crear archivo con el resultado del analizar imagen
    analisisImagen(formas, imgOutName);

    // escribo los datos en un archivo de salida
    write_pnm(imgIn, imgOutName); 

    // libero la memoria utilizada en el programa
    matrix_free(imgIn);
    for(i=0; i<ArrayList_getSize(formas); i++) {
        f = ArrayList_get(formas, i);
        Forma_free(f);
    }
    ArrayList_free(formas);
    return 0;
}


// se encarga de buscar en toda la imagen todas las formas
// que se encuentren en negro pixel=0
void buscarFormas(MATRIX* m, ArrayList* formas) {
    int i,j;
    byte** ptr = (byte**) m->ptr;
    Forma* f;

    for(i=0; i < m->rows; i++) {
        for(j=0; j < m->cols; j++) {
	        if(ptr[i][j] == 0) {
		        f = obtenerForma(m, i, j);
		        ArrayList_add(formas, f);
	        }
	    }
    }

}

// extraer forma
// se encarga de juntar todos los puntos que pertenecen a una forma
// dado un punto se encarga de buscar todos los puntos cercanos
// para conformar una forma
// durante el proceso de ir metiendo a una pila los pixeles de una forma
// esta funcion se encarga de investigar si dicho pixel es un vertice
// en cuyo caso lo almacena en la lista de vertices de la forma
Forma* obtenerForma(MATRIX* m, int i, int j) {
    Stack* s = new_Stack();
    Forma* f = new_Forma();
    byte** ptr = (byte**)m->ptr;

    Punto* p = new_Punto(i, j);
    Stack_push(s, p);

    while(!Stack_isEmpty(s)) {
        // sacamos al punto (pixel) de la pila
        p = Stack_pop(s);
        // metemos a sus vecinos que forman parte de la forma en
        // la pila
        agregarVecinos(p, s, m);
        // checamos si el punto indicado es un vertice
	    if(esVertice(p, m)) {
	        Forma_addVertice(f, p);
	    }
        // agregamos el punto a la forma
        Forma_addPoint(f, p);

        // marcamos el punto que forma parte de la forma
        ptr[p->y][p->x] = numForma;
    }
    numForma++;
    ArrayList_free(s);
    return f;
}

// se encarga de meter a los vecinos de un punto en la pila
// que formen parte de una forma
void agregarVecinos(Punto* p, Stack* s, MATRIX* m) {
    byte** ptr =(byte**) m->ptr;
    int i,j;
    int x;
    int y;

    for(i=-1; i<=1; i++) {
	    for(j=-1; j<=1; j++) {
	        x = p->x + i;
	        y = p->y + j;
	        if(y<m->rows && x<m->cols && ptr[y][x] == 0) {
		        Stack_push(s, new_Punto(y, x));
	        }
	    }
    }
}

// se encarga de crear la caja envolvente dados los puntos 
// max y min
// dibuja la caja envolvente en una image
void crearCajaEnvolvente(MATRIX* m, Punto* min, Punto* max) {
    byte** ptr =(byte**) m->ptr;
    int i,j;

    for(i=min->x; i<=max->x; i++) {
	    ptr[min->y][i] = (byte) COLOR_GRAY;
	    ptr[max->y][i] = (byte) COLOR_GRAY;
    }
    for(i=min->y; i<=max->y; i++) {
	    ptr[i][min->x] = (byte) COLOR_GRAY;
	    ptr[i][max->x] = (byte) COLOR_GRAY;
    }
}

// checa si un punto es un vertice
// para checar vertices se concideran solo 4 posibilidades de vertices
int esVertice(Punto* p, MATRIX* m) {
    return esEsquinaIzqArriba(p, m) || 
	   esEsquinaDerArriba(p, m) || 
	   esEsquinaDerAbajo(p, m) ||
	   esEsquinaIzqAbajo(p, m);
}
//  xxxx
//  xp
//  x
//  x
int esEsquinaIzqArriba(Punto* p, MATRIX* m) {
    byte** ptr = (byte**) m->ptr;
    int sum = 0;
    // arriba
    sum += ptr[p->y-1][p->x-1];
    sum += ptr[p->y-1][p->x];
    sum += ptr[p->y-1][p->x+1];
    // izquierda
    sum += ptr[p->y][p->x-1];
    sum += ptr[p->y+1][p->x-1];
    return sum == 5*255;
}

//
//  xxxx
//    px
//     x
//     x
//
int esEsquinaDerArriba(Punto* p, MATRIX* m) {
    byte** ptr = (byte**) m->ptr;
    
    int sum = 0;

    // arriba
    sum += ptr[p->y-1][p->x-1];
    sum += ptr[p->y-1][p->x];
    sum += ptr[p->y-1][p->x+1];
    // derecha
    sum += ptr[p->y][p->x+1];
    sum += ptr[p->y+1][p->x+1];

    return sum == 5*255;
}

//     x
//     x
//    px
//  xxxx
//
int esEsquinaDerAbajo(Punto* p, MATRIX* m) {
    byte** ptr = (byte**) m->ptr;
    
    int sum = 0;

    // abajo
    sum += ptr[p->y+1][p->x-1];
    sum += ptr[p->y+1][p->x];
    sum += ptr[p->y+1][p->x+1];
    // derecha
    sum += ptr[p->y-1][p->x+1];
    sum += ptr[p->y][p->x+1];

    return sum == 5*255;
}

//  x
//  x 
//  xp
//  xxxx
//
int esEsquinaIzqAbajo(Punto* p, MATRIX* m) {
    byte** ptr = (byte**) m->ptr;
    
    int sum = 0;

    // izquierda
    sum += ptr[p->y+1][p->x-1];
    sum += ptr[p->y+1][p->x];
    sum += ptr[p->y+1][p->x+1];
    // abajo
    sum += ptr[p->y-1][p->x-1];
    sum += ptr[p->y][p->x-1];

    return sum == 5*255;
}

// se encarga de dibujar un cuadrado en la ubicacion de cada vertice
void dibujarVertices(MATRIX* m, ArrayList* v) {
    int i;
    Punto* p;
    for(i=0; i<ArrayList_getSize(v); i++) {
	    p = ArrayList_get(v, i);
	    dibujarPunto(m, p, COLOR_DARK_GRAY, 2);
    }
}

// se encarga de dibujar un cuadro donde el centro es el punto p
void dibujarPunto(MATRIX* m, Punto* p, int tonoGris, int grosor) {
    byte** ptr = (byte**) m->ptr;

    int i,j;
    for(i=-grosor; i<=grosor; i++) {
	    for(j=-grosor; j<=grosor; j++) {
	        ptr[p->y+i][p->x+j] = tonoGris;
	    }
    }

}

// Angulo recto
// busca el angulo recto dados los vertices
// en un conjunto de vertices
// se buscan tercias de puntos
// y se les calcula el angulo que forman entre ellos
// en este caso nos quedamos con aquel punto que tenga el 
// producto punto mas pequeño
Punto* buscarAnguloRecto(ArrayList* v) {
    Punto* a;
    Punto* b;
    Punto* c;
    Punto* r;
    int min = 10000000;
    int prod;
    int i,j,k;
    // buscamos las permutaciones de tres vertices de entre n
    for(i=0; i<ArrayList_getSize(v); i++) {
	    for(j=0; j<ArrayList_getSize(v); j++) {
	        for(k=0; k<ArrayList_getSize(v); k++) {
		        a = ArrayList_get(v, i);
		        b = ArrayList_get(v, j);
		        c = ArrayList_get(v, k);
                // checamos que sean diferentes vertices
		        if(!Punto_equals(a,b) && !Punto_equals(a,c)) {
                    // restamos para mandarlos al origen
		            Punto* r1 = Punto_restarAbs(a,b);
		            Punto* r2 = Punto_restarAbs(a,c);
                    // se calcula el producto punto
		            prod = Punto_prodEscalar(r1,r2);
                    // nos quedamos con el menor
		            if(prod < min) {
			            min = prod;
			            r = a;
		            }
		        }
	        }
    	}
    }
    return r;
}

// dibuja numeros en una imagen solo puede dibujar
// numeros del 1 al 8
// toma como referencia un punto
void dibujarNumero(MATRIX* m, Punto* p, int n, int tonoGris) {
    byte** ptr = (byte**)m->ptr;
    int x = p->x-10;
    int y = p->y-10;

    switch(n) {
	case 1:
	    ptr[y][x+2] = tonoGris;
	    ptr[y+1][x+2] = tonoGris;
	    ptr[y+1][x+1] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+2] = tonoGris;
	    ptr[y+4][x+1] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    ptr[y+4][x+3] = tonoGris;
	    break;
	case 2:
	    ptr[y][x+1] = tonoGris;
	    ptr[y][x+2] = tonoGris;
	    ptr[y+1][x+3] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+1] = tonoGris;
	    ptr[y+4][x+1] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    ptr[y+4][x+3] = tonoGris;
	    break;
	case 3:
	    ptr[y][x+1] = tonoGris;
	    ptr[y][x+2] = tonoGris;
	    ptr[y+1][x+3] = tonoGris;
	    ptr[y+2][x+1] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+3] = tonoGris;
	    ptr[y+4][x+1] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    break;
	case 4:
	    ptr[y][x+3] = tonoGris;
	    ptr[y+1][x+1] = tonoGris;
	    ptr[y+1][x+3] = tonoGris;
	    ptr[y+2][x+1] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+2][x+3] = tonoGris;
	    ptr[y+3][x+3] = tonoGris;
	    ptr[y+4][x+3] = tonoGris;
	    break;
	case 5:
	    ptr[y][x+1] = tonoGris;
	    ptr[y][x+2] = tonoGris;
	    ptr[y+1][x+1] = tonoGris;
	    ptr[y+2][x+1] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+3] = tonoGris;
	    ptr[y+4][x+1] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    break;
	case 6:
	    ptr[y][x+2] = tonoGris;
	    ptr[y][x+3] = tonoGris;
	    ptr[y+1][x+1] = tonoGris;
	    ptr[y+2][x+1] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+1] = tonoGris;
	    ptr[y+3][x+3] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    break;
	case 7:
	    ptr[y][x+1] = tonoGris;
	    ptr[y][x+2] = tonoGris;
	    ptr[y][x+3] = tonoGris;
	    ptr[y+1][x+3] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+2] = tonoGris;
	    ptr[y+4][x+1] = tonoGris;
	    //ptr[y+4][x+2] = tonoGris;
	    break;
	case 8:
	    ptr[y][x+2] = tonoGris;
	    ptr[y+1][x+1] = tonoGris;
	    ptr[y+1][x+3] = tonoGris;
	    ptr[y+2][x+2] = tonoGris;
	    ptr[y+3][x+1] = tonoGris;
	    ptr[y+3][x+3] = tonoGris;
	    ptr[y+4][x+2] = tonoGris;
	    //ptr[y+4][x+2] = tonoGris;
	    break;
	default:
	    break;
    }
}

// funcion que genera un archivo con los datos despues de analizar la image
// indica cual fue el area de las figuras (num pixeles que la confirman)
// numero de vertices identificados
// vertice que tiene el angulo recto
void analisisImagen(ArrayList* formas, char* name) {
    char nameFile[MAX_CHARS];
    sprintf(nameFile, "%s.analisis.txt",name);
    FILE* file = fopen(nameFile, "w");

    fprintf(file, "Imagen analizada: %s\n", name);
    
    int numFormas = ArrayList_getSize(formas);
    
    fprintf(file, "Formas identificadas: %d\n", numFormas);

    int i,j;
    Forma* f;
    ArrayList* v;
    Punto* p;
    Punto* ar;
    Punto* min;
    Punto* max;
    for(i=0; i<numFormas; i++) {
        f = ArrayList_get(formas, i);
        fprintf(file, "Identificador de forma: %d\n", (i+1));

        fprintf(file, "Area: %d px\n", Forma_getNumPoints(f));

        max = Forma_getMaxPoint(f);
        min = Forma_getMinPoint(f);
        fprintf(file, "Caja envolvente: Pmin:(%d,%d), Pmax(%d,%d)\n",min->x,min->y,max->x,max->y);

        fprintf(file,"Vertices detectados:\n");

        v = Forma_getVertices(f);
        
        for(j=0; j<ArrayList_getSize(v); j++) {
            p = ArrayList_get(v, j);
            fprintf(file, "V%d = (%d,%d)\n",j,p->x,p->y);
        }

        ar = Forma_getVerticeAnguloRecto(f);
        fprintf(file,"Vertice de 90°: (%d,%d)\n\n",ar->x,ar->y);
       
    }

    fflush(file);
    fclose(file);
}



